@@ -465,6 +465,9 @@ type DiagnosticExpectation struct {
 
 	// path is the scratch workdir-relative path to the file being asserted on.
 	path string
+
+	// optionally, the diagnostic source
+	source string
 }
 
 // Check implements the Expectation interface.
@@ -489,6 +492,9 @@ func (e DiagnosticExpectation) Check(s State) Verdict {
 				continue
 			}
 		}
+		if e.source != "" && e.source != d.Source {
+			continue
+		}
 		found = true
 		break
 	}
@@ -515,6 +521,9 @@ func (e DiagnosticExpectation) Description() string {
 	if e.message != "" {
 		desc += fmt.Sprintf(" with message %q", e.message)
 	}
+	if e.source != "" {
+		desc += fmt.Sprintf(" from source %q", e.source)
+	}
 	return desc
 }
 
@@ -619,6 +628,14 @@ func (e *Env) DiagnosticAtRegexpWithMessage(name, re, msg string) DiagnosticExpe
 	return DiagnosticExpectation{path: name, pos: &pos, re: re, present: true, message: msg}
 }
 
+// DiagnosticAtRegexpFromSource expects a diagnostic at the first position
+// matching re, from the given source.
+func (e *Env) DiagnosticAtRegexpFromSource(name, re, source string) DiagnosticExpectation {
+	e.T.Helper()
+	pos := e.RegexpSearch(name, re)
+	return DiagnosticExpectation{path: name, pos: &pos, re: re, present: true, source: source}
+}
+
 // DiagnosticAt asserts that there is a diagnostic entry at the position
 // specified by line and col, for the workdir-relative path name.
 func DiagnosticAt(name string, line, col int) DiagnosticExpectation {
