@@ -7,15 +7,17 @@ package lsp
 import (
 	"bytes"
 	"context"
+	"errors"
 	"fmt"
 	"path/filepath"
-	"sync"
+	"time"
 
+	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/jsonrpc2"
 	"golang.org/x/tools/internal/lsp/protocol"
 	"golang.org/x/tools/internal/lsp/source"
 	"golang.org/x/tools/internal/span"
-	errors "golang.org/x/xerrors"
+	"golang.org/x/tools/internal/xcontext"
 )
 
 // ModificationSource identifies the originating cause of a file modification.
@@ -117,12 +119,42 @@ func (s *Server) didChange(ctx context.Context, params *protocol.DidChangeTextDo
 	if err := s.didModifyFiles(ctx, []source.FileModification{c}, FromDidChange); err != nil {
 		return err
 	}
+	return s.warnAboutModifyingGeneratedFiles(ctx, uri)
+}
 
+// warnAboutModifyingGeneratedFiles shows a warning if a user tries to edit a
+// generated file for the first time.
+func (s *Server) warnAboutModifyingGeneratedFiles(ctx context.Context, uri span.URI) error {
 	s.changedFilesMu.Lock()
-	defer s.changedFilesMu.Unlock()
+	_, ok := s.changedFiles[uri]
+	if !ok {
+		s.changedFiles[uri] = struct{}{}
+	}
+	s.changedFilesMu.Unlock()
 
-	s.changedFiles[uri] = struct{}{}
-	return nil
+	// This file has already been edited before.
+	if ok {
+		return nil
+	}
+
+	// Ideally, we should be able to specify that a generated file should
+	// be opened as read-only. Tell the user that they should not be
+	// editing a generated file.
+	view, err := s.session.ViewOf(uri)
+	if err != nil {
+		return err
+	}
+	snapshot, release := view.Snapshot(ctx)
+	isGenerated := source.IsGenerated(ctx, snapshot, uri)
+	release()
+
+	if !isGenerated {
+		return nil
+	}
+	return s.client.ShowMessage(ctx, &protocol.ShowMessageParams{
+		Message: fmt.Sprintf("Do not edit this file! %s is a generated file.", uri.Filename()),
+		Type:    protocol.Warning,
+	})
 }
 
 func (s *Server) didChangeWatchedFiles(ctx context.Context, params *protocol.DidChangeWatchedFilesParams) error {
@@ -148,9 +180,8 @@ func (s *Server) didSave(ctx context.Context, params *protocol.DidSaveTextDocume
 		return nil
 	}
 	c := source.FileModification{
-		URI:     uri,
-		Action:  source.Save,
-		Version: params.TextDocument.Version,
+		URI:    uri,
+		Action: source.Save,
 	}
 	if params.Text != nil {
 		c.Text = []byte(*params.Text)
@@ -174,128 +205,109 @@ func (s *Server) didClose(ctx context.Context, params *protocol.DidCloseTextDocu
 }
 
 func (s *Server) didModifyFiles(ctx context.Context, modifications []source.FileModification, cause ModificationSource) error {
-	// diagnosticWG tracks outstanding diagnostic work as a result of this file
-	// modification.
-	var diagnosticWG sync.WaitGroup
+	diagnoseDone := make(chan struct{})
 	if s.session.Options().VerboseWorkDoneProgress {
-		work := s.progress.start(ctx, DiagnosticWorkTitle(cause), "Calculating file diagnostics...", nil, nil)
+		work := s.progress.Start(ctx, DiagnosticWorkTitle(cause), "Calculating file diagnostics...", nil, nil)
 		defer func() {
 			go func() {
-				diagnosticWG.Wait()
-				work.end("Done.")
+				<-diagnoseDone
+				work.End("Done.")
 			}()
 		}()
 	}
-	snapshots, releases, deletions, err := s.session.DidModifyFiles(ctx, modifications)
+
+	onDisk := cause == FromDidChangeWatchedFiles
+	delay := s.session.Options().ExperimentalWatchedFileDelay
+	s.fileChangeMu.Lock()
+	defer s.fileChangeMu.Unlock()
+	if !onDisk || delay == 0 {
+		// No delay: process the modifications immediately.
+		return s.processModifications(ctx, modifications, onDisk, diagnoseDone)
+	}
+	// Debounce and batch up pending modifications from watched files.
+	pending := &pendingModificationSet{
+		diagnoseDone: diagnoseDone,
+		changes:      modifications,
+	}
+	// Invariant: changes appended to s.pendingOnDiskChanges are eventually
+	// handled in the order they arrive. This guarantee is only partially
+	// enforced here. Specifically:
+	//  1. s.fileChangesMu ensures that the append below happens in the order
+	//     notifications were received, so that the changes within each batch are
+	//     ordered properly.
+	//  2. The debounced func below holds s.fileChangesMu while processing all
+	//     changes in s.pendingOnDiskChanges, ensuring that no batches are
+	//     processed out of order.
+	//  3. Session.ExpandModificationsToDirectories and Session.DidModifyFiles
+	//     process changes in order.
+	s.pendingOnDiskChanges = append(s.pendingOnDiskChanges, pending)
+	ctx = xcontext.Detach(ctx)
+	okc := s.watchedFileDebouncer.debounce("", 0, time.After(delay))
+	go func() {
+		if ok := <-okc; !ok {
+			return
+		}
+		s.fileChangeMu.Lock()
+		var allChanges []source.FileModification
+		// For accurate progress notifications, we must notify all goroutines
+		// waiting for the diagnose pass following a didChangeWatchedFiles
+		// notification. This is necessary for regtest assertions.
+		var dones []chan struct{}
+		for _, pending := range s.pendingOnDiskChanges {
+			allChanges = append(allChanges, pending.changes...)
+			dones = append(dones, pending.diagnoseDone)
+		}
+
+		allDone := make(chan struct{})
+		if err := s.processModifications(ctx, allChanges, onDisk, allDone); err != nil {
+			event.Error(ctx, "processing delayed file changes", err)
+		}
+		s.pendingOnDiskChanges = nil
+		s.fileChangeMu.Unlock()
+		<-allDone
+		for _, done := range dones {
+			close(done)
+		}
+	}()
+	return nil
+}
+
+// processModifications update server state to reflect file changes, and
+// triggers diagnostics to run asynchronously. The diagnoseDone channel will be
+// closed once diagnostics complete.
+func (s *Server) processModifications(ctx context.Context, modifications []source.FileModification, onDisk bool, diagnoseDone chan struct{}) error {
+	s.stateMu.Lock()
+	if s.state >= serverShutDown {
+		// This state check does not prevent races below, and exists only to
+		// produce a better error message. The actual race to the cache should be
+		// guarded by Session.viewMu.
+		s.stateMu.Unlock()
+		close(diagnoseDone)
+		return errors.New("server is shut down")
+	}
+	s.stateMu.Unlock()
+	// If the set of changes included directories, expand those directories
+	// to their files.
+	modifications = s.session.ExpandModificationsToDirectories(ctx, modifications)
+
+	snapshots, releases, err := s.session.DidModifyFiles(ctx, modifications)
 	if err != nil {
+		close(diagnoseDone)
 		return err
 	}
 
-	for _, uri := range deletions {
-		if err := s.client.PublishDiagnostics(ctx, &protocol.PublishDiagnosticsParams{
-			URI:         protocol.URIFromSpanURI(uri),
-			Diagnostics: []protocol.Diagnostic{},
-			Version:     0,
-		}); err != nil {
-			return err
-		}
-	}
-	snapshotByURI := make(map[span.URI]source.Snapshot)
-	for _, c := range modifications {
-		snapshotByURI[c.URI] = nil
-	}
-	// Avoid diagnosing the same snapshot twice.
-	snapshotSet := make(map[source.Snapshot][]span.URI)
-	for uri := range snapshotByURI {
-		view, err := s.session.ViewOf(uri)
-		if err != nil {
-			return err
-		}
-		var snapshot source.Snapshot
-		for _, s := range snapshots {
-			if s.View() == view {
-				if snapshot != nil {
-					return errors.Errorf("duplicate snapshots for the same view")
-				}
-				snapshot = s
-			}
-		}
-		// If the file isn't in any known views (for example, if it's in a dependency),
-		// we may not have a snapshot to map it to. As a result, we won't try to
-		// diagnose it. TODO(rstambler): Figure out how to handle this better.
-		if snapshot == nil {
-			continue
-		}
-		snapshotSet[snapshot] = append(snapshotSet[snapshot], uri)
-		snapshotByURI[uri] = snapshot
-	}
-
-	for _, mod := range modifications {
-		if mod.OnDisk || mod.Action != source.Change {
-			continue
-		}
-		snapshot, ok := snapshotByURI[mod.URI]
-		if !ok {
-			continue
-		}
-		// Ideally, we should be able to specify that a generated file should be opened as read-only.
-		// Tell the user that they should not be editing a generated file.
-		if s.wasFirstChange(mod.URI) && source.IsGenerated(ctx, snapshot, mod.URI) {
-			if err := s.client.ShowMessage(ctx, &protocol.ShowMessageParams{
-				Message: fmt.Sprintf("Do not edit this file! %s is a generated file.", mod.URI.Filename()),
-				Type:    protocol.Warning,
-			}); err != nil {
-				return err
-			}
-		}
-	}
-
-	for snapshot, uris := range snapshotSet {
-		// If a modification comes in for the view's go.mod file and the view
-		// was never properly initialized, or the view does not have
-		// a go.mod file, try to recreate the associated view.
-		if modfile := snapshot.View().ModFile(); modfile == "" {
-			for _, uri := range uris {
-				// Don't rebuild the view until the go.mod is on disk.
-				if !snapshot.IsSaved(uri) {
-					continue
-				}
-				fh, err := snapshot.GetFile(ctx, uri)
-				if err != nil {
-					return err
-				}
-				switch fh.Kind() {
-				case source.Mod:
-					newSnapshot, release, err := snapshot.View().Rebuild(ctx)
-					releases = append(releases, release)
-					if err != nil {
-						return err
-					}
-					// Update the snapshot to the rebuilt one.
-					snapshot = newSnapshot
-				}
-			}
-		}
-		diagnosticWG.Add(1)
-		go func(snapshot source.Snapshot, uris []span.URI) {
-			defer diagnosticWG.Done()
-			s.diagnoseSnapshot(snapshot)
-		}(snapshot, uris)
-	}
-
 	go func() {
-		diagnosticWG.Wait()
+		s.diagnoseSnapshots(snapshots, onDisk)
 		for _, release := range releases {
 			release()
 		}
+		close(diagnoseDone)
 	}()
+
 	// After any file modifications, we need to update our watched files,
 	// in case something changed. Compute the new set of directories to watch,
 	// and if it differs from the current set, send updated registrations.
-	if err := s.updateWatchedDirectories(ctx, snapshots); err != nil {
-		return err
-	}
-	return nil
+	return s.updateWatchedDirectories(ctx)
 }
 
 // DiagnosticWorkTitle returns the title of the diagnostic work resulting from a
@@ -304,20 +316,9 @@ func DiagnosticWorkTitle(cause ModificationSource) string {
 	return fmt.Sprintf("diagnosing %v", cause)
 }
 
-func (s *Server) wasFirstChange(uri span.URI) bool {
-	s.changedFilesMu.Lock()
-	defer s.changedFilesMu.Unlock()
-
-	if s.changedFiles == nil {
-		s.changedFiles = make(map[span.URI]struct{})
-	}
-	_, ok := s.changedFiles[uri]
-	return !ok
-}
-
 func (s *Server) changedText(ctx context.Context, uri span.URI, changes []protocol.TextDocumentContentChangeEvent) ([]byte, error) {
 	if len(changes) == 0 {
-		return nil, errors.Errorf("%w: no content changes provided", jsonrpc2.ErrInternal)
+		return nil, fmt.Errorf("%w: no content changes provided", jsonrpc2.ErrInternal)
 	}
 
 	// Check if the client sent the full content of the file.
@@ -335,7 +336,7 @@ func (s *Server) applyIncrementalChanges(ctx context.Context, uri span.URI, chan
 	}
 	content, err := fh.Read()
 	if err != nil {
-		return nil, errors.Errorf("%w: file not found (%v)", jsonrpc2.ErrInternal, err)
+		return nil, fmt.Errorf("%w: file not found (%v)", jsonrpc2.ErrInternal, err)
 	}
 	for _, change := range changes {
 		// Make sure to update column mapper along with the content.
@@ -346,18 +347,18 @@ func (s *Server) applyIncrementalChanges(ctx context.Context, uri span.URI, chan
 			Content:   content,
 		}
 		if change.Range == nil {
-			return nil, errors.Errorf("%w: unexpected nil range for change", jsonrpc2.ErrInternal)
+			return nil, fmt.Errorf("%w: unexpected nil range for change", jsonrpc2.ErrInternal)
 		}
 		spn, err := m.RangeSpan(*change.Range)
 		if err != nil {
 			return nil, err
 		}
 		if !spn.HasOffset() {
-			return nil, errors.Errorf("%w: invalid range for content change", jsonrpc2.ErrInternal)
+			return nil, fmt.Errorf("%w: invalid range for content change", jsonrpc2.ErrInternal)
 		}
 		start, end := spn.Start().Offset(), spn.End().Offset()
 		if end < start {
-			return nil, errors.Errorf("%w: invalid range for content change", jsonrpc2.ErrInternal)
+			return nil, fmt.Errorf("%w: invalid range for content change", jsonrpc2.ErrInternal)
 		}
 		var buf bytes.Buffer
 		buf.Write(content[:start])
